print("import: 'owslib.coverage'")
import owslib.coverage

print("import: 'owslib.waterml'")
import owslib.waterml

print("import: 'owslib.swe.observation'")
import owslib.swe.observation

print("import: 'owslib.swe.sensor'")
import owslib.swe.sensor

print("import: 'owslib'")
import owslib

print("import: 'owslib.feature'")
import owslib.feature

print("import: 'owslib.swe'")
import owslib.swe

